#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <conio.h>
#include <dos.h>
#include "lfb.h"

int SCREEN_WIDTH  = 1024;
int SCREEN_HEIGHT  = 768;


char *vram = NULL;

/*---------------------------- Global Variables ---------------------------*/
int     VESABuf_len = 1024;         /* Length of VESABuf                */
int     VESABuf_sel = 0;            /* Selector for VESABuf             */
int     VESABuf_rseg;               /* Real mode segment of VESABuf     */
short   modeList[50];               /* List of available VBE modes      */
int     xres,yres;                  /* Video mode resolution            */
int     bytesperline;               /* Bytes per scanline for mode      */
long    imageSize;                  /* Length of the video image        */
char    far *LFBPtr;                /* Pointer to linear framebuffer    */

/*------------------------- DPMI interface routines -----------------------*/

/****************************************************************************
* Function:     DPMI_allocRealSeg
* Parameters:   size    - Size of memory block to allocate
*               sel     - Place to return protected mode selector
*               r_seg   - Place to return real mode segment
* Description:  Allocates a block of real mode memory using DPMI services.
*               This routine returns both a protected mode selector and
*               real mode segment for accessing the memory block.
****************************************************************************/
void DPMI_allocRealSeg(int size,int *sel,int *r_seg)
{
    union REGS      r;

    r.w.ax = 0x100;                 /* DPMI allocate DOS memory         */
    r.w.bx = (size + 0xF) >> 4;     /* number of paragraphs             */
    int386(0x31, &r, &r);

    if (r.w.cflag)
        printf("DPMI_allocRealSeg failed!");

    *sel = r.w.dx;                  /* Protected mode selector          */
    *r_seg = r.w.ax;                /* Real mode segment                */
}

/****************************************************************************
* Function:     DPMI_allocRealSeg
* Parameters:   sel - Protected mode selector of block to free
* Description:  Frees a block of real mode memory.
****************************************************************************/
void DPMI_freeRealSeg(unsigned sel)
{
    union REGS  r;

    r.w.ax = 0x101;                 /* DPMI free DOS memory             */
    r.w.dx = sel;                   /* DX := selector from 0x100        */
    int386(0x31, &r, &r);
}

typedef struct {
    long    edi;
    long    esi;
    long    ebp;
    long    reserved;
    long    ebx;
    long    edx;
    long    ecx;
    long    eax;
    short   flags;
    short   es,ds,fs,gs,ip,cs,sp,ss;
    } _RMREGS;

#define IN(reg)     rmregs.e##reg = in->x.reg
#define OUT(reg)    out->x.reg = rmregs.e##reg

/****************************************************************************
* Function:     DPMI_int86
* Parameters:   intno   - Interrupt number to issue
*               in      - Pointer to structure for input registers
*               out     - Pointer to structure for output registers
* Returns:      Value returned by interrupt in AX
* Description:  Issues a real mode interrupt using DPMI services.
****************************************************************************/
int DPMI_int86(int intno, RMREGS *in, RMREGS *out)
{
    _RMREGS         rmregs;
    union REGS      r;
    struct SREGS    sr;

    memset(&rmregs, 0, sizeof(rmregs));
    IN(ax); IN(bx); IN(cx); IN(dx); IN(si); IN(di);

    segread(&sr);
    r.w.ax = 0x300;                 /* DPMI issue real interrupt        */
    r.h.bl = intno;
    r.h.bh = 0;
    r.w.cx = 0;
    sr.es = sr.ds;
    r.x.edi = (unsigned)&rmregs;
    int386x(0x31, &r, &r, &sr);     /* Issue the interrupt              */

    OUT(ax); OUT(bx); OUT(cx); OUT(dx); OUT(si); OUT(di);
    out->x.cflag = rmregs.flags & 0x1;
    return out->x.ax;
}


/****************************************************************************
* Function:     DPMI_int86
* Parameters:   intno   - Interrupt number to issue
*               in      - Pointer to structure for input registers
*               out     - Pointer to structure for output registers
*               sregs   - Values to load into segment registers
* Returns:      Value returned by interrupt in AX
* Description:  Issues a real mode interrupt using DPMI services.
****************************************************************************/
int DPMI_int86x(int intno, RMREGS *in, RMREGS *out, RMSREGS *sregs)
{
    _RMREGS         rmregs;
    union REGS      r;
    struct SREGS    sr;

    memset(&rmregs, 0, sizeof(rmregs));
    IN(ax); IN(bx); IN(cx); IN(dx); IN(si); IN(di);
    rmregs.es = sregs->es;
    rmregs.ds = sregs->ds;

    segread(&sr);
    r.w.ax = 0x300;                 /* DPMI issue real interrupt        */
    r.h.bl = intno;
    r.h.bh = 0;
    r.w.cx = 0;
    sr.es = sr.ds;
    r.x.edi = (unsigned)&rmregs;
    int386x(0x31, &r, &r, &sr);     /* Issue the interrupt */

    OUT(ax); OUT(bx); OUT(cx); OUT(dx); OUT(si); OUT(di);
    sregs->es = rmregs.es;
    sregs->cs = rmregs.cs;
    sregs->ss = rmregs.ss;
    sregs->ds = rmregs.ds;
    out->x.cflag = rmregs.flags & 0x1;
    return out->x.ax;
}


/****************************************************************************
* Function:     DPMI_allocSelector
* Returns:      Newly allocated protected mode selector
* Description:  Allocates a new protected mode selector using DPMI
*               services. This selector has a base address and limit of 0.
****************************************************************************/
int DPMI_allocSelector(void)
{
    int         sel;
    union REGS  r;

    r.w.ax = 0;                     /* DPMI allocate selector           */
    r.w.cx = 1;                     /* Allocate a single selector       */
    int386(0x31, &r, &r);
    if (r.x.cflag)
        printf("DPMI_allocSelector() failed!");
    sel = r.w.ax;

    r.w.ax = 9;                     /* DPMI set access rights           */
    r.w.bx = sel;
    r.w.cx = 0x8092;                /* 32 bit page granular             */
    int386(0x31, &r, &r);
    return sel;
}


/****************************************************************************
* Function:     DPMI_mapPhysicalToLinear
* Parameters:   physAddr    - Physical memory address to map
*               limit       - Length-1 of physical memory region to map
* Returns:      Starting linear address for mapped memory
* Description:  Maps a section of physical memory into the linear address
*               space of a process using DPMI calls. Note that this linear
*               address cannot be used directly, but must be used as the
*               base address for a selector.
****************************************************************************/
long DPMI_mapPhysicalToLinear(long physAddr,long limit)
{
    union REGS  r;

    r.w.ax = 0x800;                 /* DPMI map physical to linear      */
    r.w.bx = physAddr >> 16;
    r.w.cx = physAddr & 0xFFFF;
    r.w.si = limit >> 16;
    r.w.di = limit & 0xFFFF;
    int386(0x31, &r, &r);
    if (r.x.cflag)
        printf("DPMI_mapPhysicalToLinear() failed!");
    return ((long)r.w.bx << 16) + r.w.cx;
}


/****************************************************************************
* Function:     DPMI_setSelectorBase
* Parameters:   sel     - Selector to change base address for
*               linAddr - Linear address used for new base address
* Description:  Sets the base address for the specified selector.
****************************************************************************/
void DPMI_setSelectorBase(int sel,long linAddr)
{
    union REGS  r;

    r.w.ax = 7;                     /* DPMI set selector base address   */
    r.w.bx = sel;
    r.w.cx = linAddr >> 16;
    r.w.dx = linAddr & 0xFFFF;
    int386(0x31, &r, &r);
    if (r.x.cflag)
        printf("DPMI_setSelectorBase() failed!");
}


/****************************************************************************
* Function:     DPMI_setSelectorLimit
* Parameters:   sel     - Selector to change limit for
*               limit   - Limit-1 for the selector
* Description:  Sets the memory limit for the specified selector.
****************************************************************************/
void DPMI_setSelectorLimit(int sel,long limit)
{
    union REGS  r;

    r.w.ax = 8;                     /* DPMI set selector limit          */
    r.w.bx = sel;
    r.w.cx = limit >> 16;
    r.w.dx = limit & 0xFFFF;
    int386(0x31, &r, &r);
    if (r.x.cflag)
        printf("DPMI_setSelectorLimit() failed!");
}

/*-------------------------- VBE Interface routines -----------------------*/
static void ExitVBEBuf(void)
{
    DPMI_freeRealSeg(VESABuf_sel);
}


/****************************************************************************
* Function:     VBE_initRMBuf
* Description:  Initialises the VBE transfer buffer in real mode memory.
*               This routine is called by the VESAVBE module every time
*               it needs to use the transfer buffer, so we simply allocate
*               it once and then return.
****************************************************************************/
void VBE_initRMBuf(void)
{
    if (!VESABuf_sel)
    {
        DPMI_allocRealSeg(VESABuf_len, &VESABuf_sel, &VESABuf_rseg);
        atexit(ExitVBEBuf);
    }
}


/****************************************************************************
* Function:     VBE_callESDI
* Parameters:   regs    - Registers to load when calling VBE
*               buffer  - Buffer to copy VBE info block to
*               size    - Size of buffer to fill
* Description:  Calls the VESA VBE and passes in a buffer for the VBE to
*               store information in, which is then copied into the users
*               buffer space. This works in protected mode as the buffer
*               passed to the VESA VBE is allocated in conventional
*               memory, and is then copied into the users memory block.
****************************************************************************/
void VBE_callESDI(RMREGS *regs, void *buffer, int size)
{
    RMSREGS sregs;

    VBE_initRMBuf();
    sregs.es = VESABuf_rseg;
    regs->x.di = 0;
    _fmemcpy(MK_FP(VESABuf_sel,0),buffer,size);
    DPMI_int86x(0x10, regs, regs, &sregs);
    _fmemcpy(buffer,MK_FP(VESABuf_sel,0),size);
}


/****************************************************************************
* Function:     VBE_detect
* Parameters:   vgaInfo - Place to store the VGA information block
* Returns:      VBE version number, or 0 if not detected.
* Description:  Detects if a VESA VBE is out there and functioning
*               correctly. If we detect a VBE interface we return the
*               VGAInfoBlock returned by the VBE and the VBE version number.
****************************************************************************/
int VBE_detect(void)
{
    RMREGS      regs;
    short       *p1,*p2;
    VBE_vgaInfo vgaInfo;
    /* Put 'VBE2' into the signature area so that the VBE 2.0 BIOS knows
     * that we have passed a 512 byte extended block to it, and wish
     * the extended information to be filled in.
     */
    strncpy(vgaInfo.VESASignature,"VBE2",4);
    /* Get the SuperVGA Information block */
    regs.x.ax = 0x4F00;
    VBE_callESDI(&regs, &vgaInfo, sizeof(VBE_vgaInfo));
    if (regs.x.ax != 0x004F)
        return 0;
    if (strncmp(vgaInfo.VESASignature,"VESA",4) != 0)
        return 0;
    /* Now that we have detected a VBE interface, copy the list of available
     * video modes into our local buffer. We *must* copy this mode list, since
     * the VBE will build the mode list in the VBE_vgaInfo buffer that we have
     * passed, so the next call to the VBE will trash the list of modes. */
    p1 = LfbMapRealPointer(vgaInfo.VideoModePtr);
    p2 = modeList;
    while (*p1 != -1)
        *p2++ = *p1++;
    *p2 = -1;
    return vgaInfo.VESAVersion;
}



/****************************************************************************
* Function:     VBE_getModeInfo
* Parameters:   mode        - VBE mode to get information for
*               modeInfo    - Place to store VBE mode information
* Returns:      1 on success, 0 if function failed.
* Description:  Obtains information about a specific video mode from the
*               VBE. You should use this function to find the video mode
*               you wish to set, as the new VBE 2.0 mode numbers may be
*               completely arbitrary.
****************************************************************************/
int VBE_getModeInfo(int mode,VBE_modeInfo *modeInfo)
{
    RMREGS  regs;

    regs.x.ax = 0x4F01;             /* Get mode information         */
    regs.x.cx = mode;
    VBE_callESDI(&regs, modeInfo, sizeof(VBE_modeInfo));
    if (regs.x.ax != 0x004F)
        return 0;
    if ((modeInfo->ModeAttributes & vbeMdAvailable) == 0)
        return 0;
    return 1;
}


/****************************************************************************
* Function:     VBE_setVideoMode
* Parameters:   mode    - VBE mode number to initialise
****************************************************************************/
void VBE_setVideoMode(int mode)
{
    RMREGS  regs;
    regs.x.ax = 0x4F02;
    regs.x.bx = mode;
    DPMI_int86(0x10,&regs,&regs);
}



/*-------------------- Application specific routines ----------------------*/



/****************************************************************************
* Function:     GetPtrToLFB
* Parameters:   physAddr    - Physical memory address of linear framebuffer
* Returns:      Far pointer to the linear framebuffer memory
****************************************************************************/
void far *GetPtrToLFB(long physAddr)
{
    int     sel;
    long    linAddr,limit = (4096 * 1024) - 1;

    sel = DPMI_allocSelector();
    linAddr = DPMI_mapPhysicalToLinear(physAddr,limit);
    DPMI_setSelectorBase(sel,linAddr);
    DPMI_setSelectorLimit(sel,limit);
    return MK_FP(sel,0);
}


/****************************************************************************
* Function:     AvailableModes
* Description:  Display a list of available LFB mode resolutions.
****************************************************************************/
void AvailableModes(void)
{
    short           *p;
    VBE_modeInfo    modeInfo;

    printf("Usage: LFBPROF <xres> <yres>\n\n");
    printf("Available 256 color video modes:\n");
    for (p = modeList; *p != -1; p++) {
        if (VBE_getModeInfo(*p, &modeInfo)) {
            /* Filter out only 8 bit linear framebuffer modes */
            if ((modeInfo.ModeAttributes & vbeMdLinear) == 0)
                continue;
            if (modeInfo.MemoryModel != vbeMemPK
                || modeInfo.BitsPerPixel != 8
                || modeInfo.NumberOfPlanes != 1)
                continue;
            printf("    %4d x %4d %d bits per pixel\n",
                modeInfo.XResolution, modeInfo.YResolution,
                modeInfo.BitsPerPixel);
            }
        }
    exit(1);
}


/****************************************************************************
* Function:     InitGraphics
* Parameters:   x,y - Requested video mode resolution
* Description:  Initialise the specified video mode. We search through
*               the list of available video modes for one that matches
*               the resolution and color depth are are looking for.
****************************************************************************/
void InitGraphics(int x,int y)
{
    short           *p;
    VBE_modeInfo    modeInfo;

    for (p = modeList; *p != -1; p++) {
        if (VBE_getModeInfo(*p, &modeInfo)) {
            /* Filter out only 8 bit linear framebuffer modes */
            if ((modeInfo.ModeAttributes & vbeMdLinear) == 0)
                continue;
            if (modeInfo.MemoryModel != vbeMemPK
                || modeInfo.BitsPerPixel != 8
                || modeInfo.NumberOfPlanes != 1)
                continue;
            if (modeInfo.XResolution != x || modeInfo.YResolution != y)
                continue;
            xres = x;
            yres = y;
            bytesperline = modeInfo.BytesPerScanLine;
            imageSize = bytesperline * yres;
            VBE_setVideoMode(*p | vbeUseLFB);
            LFBPtr = GetPtrToLFB(modeInfo.PhysBasePtr);
            return;
            }
        }
    printf("Valid video mode not found\n");
    exit(1);
}

/****************************************************************************
* Function:     EndGraphics
* Description:  Restores text mode.
****************************************************************************/
void EndGraphics(void)
{
    RMREGS  regs;
    regs.x.ax = 0x3;
    DPMI_int86(0x10, &regs, &regs);
}


// mouse cursors made with 2L8 Cursor editor
// downloaded from http://ftp.gnome.org/mirror/archive/ftp.sunet.se/pub/simtelnet/msdos/mouse/

int POINT[32] = {
        0x3fff, 0x1fff, 0x0fff, 0x07ff, 0x03ff, 0x01ff, 0x00ff, 0x007f,
        0x003f, 0x003f, 0x01ff, 0x10ff, 0x30ff, 0xf87f, 0xf87f, 0xfc7f,
        0x0000, 0x4000, 0x6000, 0x7000, 0x7800, 0x7c00, 0x7e00, 0x7f00,
        0x7f80, 0x7c00, 0x6c00, 0x4600, 0x0600, 0x0300, 0x0300, 0x0000
};
/*
short Mouse_Init (void)
{
        union REGS in_reg, out_reg;
        struct SREGS segreg;

        segread(&segreg);
        in_reg.x.ax = (unsigned char)0x0;
        int86x(0x33, &in_reg, &out_reg, &segreg);

        // return # of mouse buttons (-1 if no mouse)
        return out_reg.x.bx;
}

void Mouse_Show (void)
{
        union REGS in_reg, out_reg;
        struct SREGS segreg;

        segread(&segreg);
        in_reg.x.ax = (unsigned char)0x1;
        int86x(0x33, &in_reg, &out_reg, &segreg);
}

void Mouse_Hide (void)
{
        union REGS in_reg, out_reg;
        struct SREGS segreg;

        segread(&segreg);
        in_reg.x.ax = (unsigned char)0x2;
        int86x(0x33, &in_reg, &out_reg, &segreg);
}

void Mouse_Status( int *x, int *y, int *b )
{
        union REGS in_reg, out_reg;
        struct SREGS segreg;

        segread(&segreg);
        in_reg.x.ax = 0x3;
        in_reg.x.bx = 0;
        in_reg.x.cx = 0;
        in_reg.x.dx = 0;

        int86x(0x33, &in_reg, &out_reg, &segreg);
        *b = out_reg.x.bx;
        *x = out_reg.x.cx;
        *y = out_reg.x.dx;
}


void Mouse_Cursor(int *shape, int x, int y)
{
        union REGS inreg, outreg;
        struct SREGS s;

        inreg.x.ax = 0x9;
        inreg.x.bx = x;
        inreg.x.cx = y;
        inreg.x.dx = (unsigned)shape;
        segread(&s);
        s.es = s.ds;
        int86x(0x33,&inreg,&outreg,&s);
}


int GetKey()
{
        union REGS in, out;

        in.h.ah = 0x08;
        int86(0x21, &in, &out);

        if (out.h.al == 0)
        {
                return GetKey() + 0x100;
        }
        else
        {
                return out.h.al;
        }
}
*/




void clear(unsigned char color)
{
    memset(vram, 0, SCREEN_WIDTH * SCREEN_HEIGHT);
}

void swap()
{
        LfbMemcpy(FP_SEG(LFBPtr), 0, vram, imageSize);
}


unsigned char car_sprite[] = {21, 0, 6, 0,
    0, 0, 0, 0, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 4, 4, 0, 0, 0, 0, 4, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0,
    4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 0,
    4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4, 4,
    0, 4, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 0, 0, 4, 0, 0,
    0, 0, 4, 4, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4, 4, 0, 0, 0
};

unsigned char ball_sprite[] = {5, 0, 5, 0,        /* width 5, height 5 */
      0,  11,  11,  11,   0,
     11,  14,   3,   3,   9,
     11,   3,   3,   3,   9,
     11,   3,   3,   3,   9,
      0,   9,   9,   9,   0
};


void put_pixel(int x, int y, char color)
{
    vram[x + y * SCREEN_WIDTH] = color;
}

void put_sprite (int x, int y, unsigned char sprite[])
{
        int x_count, y_count;
        int width, height;

        /* Extract the width from the first two bytes of the array, and the
                height from the third and fourth bytes of the array: */
        width = sprite[0] + (sprite[1] << 8);
        height = sprite[2] + (sprite[3] << 8);

        for (y_count = 0; y_count <= height - 1; y_count++)
        {
                for (x_count = 0; x_count <= width - 1; x_count++)
                {
                        put_pixel (x + x_count, y + y_count, sprite[4 + (y_count * width) + x_count]);
                }
        }
}

void PutSprite (int x, int y, unsigned char sprite[])
{
    int x_count, y_count;
    int width, height;

    /* Extract the width from the first two bytes of the array, and the
       height from the third and fourth bytes of the array: */
    width = sprite[0] + (sprite[1] << 8);
    height = sprite[2] + (sprite[3] << 8);

    /* Loop through the rows: */
    for (y_count = 0; y_count <= height - 1; y_count++)
        /* Loop through the columns: */
        for (x_count = 0; x_count <= width - 1; x_count++)
            put_pixel(x + x_count, y + y_count,
                sprite[4 + (y_count * width) + x_count]);
}



int init();
int render(char *data, int width, int height, int time);

int render2()
{
        int j = 0;
        int x = 0;
        int y = 0;

        while (j < 60 * 10)
        {
                clear(0);
                j++;
                PutSprite(50 + x, 100 + x, car_sprite);
                PutSprite(150 + y, 100 + 2 * y, ball_sprite);
                x++;
                y++;
                delay(15);
                swap();
        }
        return 0;
}


int main()
{
	int count = 0;

	printf("DOS3D Initializing buffers\n");
	if ( init() != 0 )
		return -1;

	printf("Entering graphics mode in two seconds...\n");
	delay(2000);

        if (VBE_detect() < 0x200)
        {
            printf("This program requires VBE 2.0; Please install UniVBE 5.1.");
        }

        InitGraphics(SCREEN_WIDTH, SCREEN_HEIGHT);



        while (count < 1000)
        {
		render(vram, SCREEN_WIDTH, SCREEN_HEIGHT, count);
		//render2();
		count++;
        }       
        EndGraphics();
        return 0;
     
}



